// Licensed to the .NET Foundation under one or more agreements.
// The .NET Foundation licenses this file to you under the MIT license.

using Microsoft.EntityFrameworkCore.Metadata.Internal;

#nullable enable

namespace Microsoft.EntityFrameworkCore.Metadata.Conventions;

/// <summary>
///     A convention that creates an optimized copy of the mutable model. This convention is typically
///     implemented by database providers to update provider annotations when creating a read-only model.
/// </summary>
/// <remarks>
///     See <see href="https://aka.ms/efcore-docs-conventions">Model building conventions</see> for more information and examples.
/// </remarks>
public class RelationalRuntimeModelConvention : RuntimeModelConvention
{
    /// <summary>
    ///     Creates a new instance of <see cref="RelationalRuntimeModelConvention" />.
    /// </summary>
    /// <param name="dependencies">Parameter object containing dependencies for this convention.</param>
    /// <param name="relationalDependencies"> Parameter object containing relational dependencies for this convention.</param>
    public RelationalRuntimeModelConvention(
        ProviderConventionSetBuilderDependencies dependencies,
        RelationalConventionSetBuilderDependencies relationalDependencies)
        : base(dependencies)
    {
        RelationalDependencies = relationalDependencies;
    }

    /// <summary>
    ///     Relational provider-specific dependencies for this service.
    /// </summary>
    protected virtual RelationalConventionSetBuilderDependencies RelationalDependencies { get; }

    /// <summary>
    ///     Updates the model annotations that will be set on the read-only object.
    /// </summary>
    /// <param name="annotations">The annotations to be processed.</param>
    /// <param name="model">The source model.</param>
    /// <param name="runtimeModel">The target model that will contain the annotations.</param>
    /// <param name="runtime">Indicates whether the given annotations are runtime annotations.</param>
    protected override void ProcessModelAnnotations(
        Dictionary<string, object?> annotations,
        IModel model,
        RuntimeModel runtimeModel,
        bool runtime)
    {
        base.ProcessModelAnnotations(annotations, model, runtimeModel, runtime);

        if (runtime)
        {
            annotations[RelationalAnnotationNames.RelationalModel] =
                RelationalModel.Create(runtimeModel, RelationalDependencies.RelationalAnnotationProvider, designTime: false);
        }
        else
        {
            annotations.Remove(RelationalAnnotationNames.Collation);

            if (annotations.TryGetValue(RelationalAnnotationNames.DbFunctions, out var functions))
            {
                var runtimeFunctions = new SortedDictionary<string, IDbFunction>(StringComparer.Ordinal);
                foreach (var (key, dbFunction) in (SortedDictionary<string, IDbFunction>)functions!)
                {
                    var runtimeFunction = Create(dbFunction, runtimeModel);
                    runtimeFunctions[key] = runtimeFunction;

                    foreach (var parameter in dbFunction.Parameters)
                    {
                        var runtimeParameter = Create(parameter, runtimeFunction);

                        CreateAnnotations(
                            parameter, runtimeParameter, static (convention, annotations, source, target, runtime) =>
                                convention.ProcessFunctionParameterAnnotations(annotations, source, target, runtime));
                    }

                    CreateAnnotations(
                        dbFunction, runtimeFunction, static (convention, annotations, source, target, runtime) =>
                            convention.ProcessFunctionAnnotations(annotations, source, target, runtime));
                }

                annotations[RelationalAnnotationNames.DbFunctions] = runtimeFunctions;
            }

            if (annotations.TryGetValue(RelationalAnnotationNames.Sequences, out var sequences))
            {
                var runtimeSequences = new SortedDictionary<(string, string?), ISequence>();
                foreach (var (key, value) in (SortedDictionary<(string, string?), ISequence>)sequences!)
                {
                    var runtimeSequence = Create(value, runtimeModel);
                    runtimeSequences[key] = runtimeSequence;

                    CreateAnnotations(
                        value, runtimeSequence, static (convention, annotations, source, target, runtime) =>
                            convention.ProcessSequenceAnnotations(annotations, source, target, runtime));
                }

                annotations[RelationalAnnotationNames.Sequences] = runtimeSequences;
            }
        }
    }

    /// <summary>
    ///     Updates the entity type annotations that will be set on the read-only object.
    /// </summary>
    /// <param name="annotations">The annotations to be processed.</param>
    /// <param name="entityType">The source entity type.</param>
    /// <param name="runtimeEntityType">The target entity type that will contain the annotations.</param>
    /// <param name="runtime">Indicates whether the given annotations are runtime annotations.</param>
    protected override void ProcessEntityTypeAnnotations(
        IDictionary<string, object?> annotations,
        IEntityType entityType,
        RuntimeEntityType runtimeEntityType,
        bool runtime)
    {
        base.ProcessEntityTypeAnnotations(annotations, entityType, runtimeEntityType, runtime);

        if (runtime)
        {
            annotations.Remove(RelationalAnnotationNames.TableMappings);
            annotations.Remove(RelationalAnnotationNames.ViewMappings);
            annotations.Remove(RelationalAnnotationNames.SqlQueryMappings);
            annotations.Remove(RelationalAnnotationNames.FunctionMappings);
            annotations.Remove(RelationalAnnotationNames.DefaultMappings);
        }
        else
        {
            annotations.Remove(RelationalAnnotationNames.CheckConstraints);
            annotations.Remove(RelationalAnnotationNames.Comment);
            annotations.Remove(RelationalAnnotationNames.IsTableExcludedFromMigrations);

            // These need to be set explicitly to prevent default values from being generated
            annotations[RelationalAnnotationNames.TableName] = entityType.GetTableName();
            annotations[RelationalAnnotationNames.Schema] = entityType.GetSchema();
            annotations[RelationalAnnotationNames.ViewName] = entityType.GetViewName();
            annotations[RelationalAnnotationNames.ViewSchema] = entityType.GetViewSchema();
            annotations[RelationalAnnotationNames.SqlQuery] = entityType.GetSqlQuery();
            annotations[RelationalAnnotationNames.FunctionName] = entityType.GetFunctionName();

            if (annotations.TryGetValue(RelationalAnnotationNames.Triggers, out var triggers))
            {
                var runtimeTriggers = new SortedDictionary<string, ITrigger>(StringComparer.Ordinal);
                foreach (var (key, trigger) in (SortedDictionary<string, ITrigger>)triggers!)
                {
                    var runtimeTrigger = Create(trigger, runtimeEntityType);
                    runtimeTriggers[key] = runtimeTrigger;

                    CreateAnnotations(
                        trigger, runtimeTrigger,
                        static (convention, annotations, source, target, runtime)
                            => convention.ProcessTriggerAnnotations(annotations, source, target, runtime));
                }

                annotations[RelationalAnnotationNames.Triggers] = runtimeTriggers;
            }
        }
    }

    private void CreateAnnotations<TSource, TTarget>(
        TSource source,
        TTarget target,
        Action<RelationalRuntimeModelConvention, Dictionary<string, object?>, TSource, TTarget, bool> process)
        where TSource : IAnnotatable
        where TTarget : AnnotatableBase
    {
        var annotations = source.GetAnnotations().ToDictionary(a => a.Name, a => a.Value);
        process(this, annotations, source, target, false);
        target.AddAnnotations(annotations);

        annotations = source.GetRuntimeAnnotations().ToDictionary(a => a.Name, a => a.Value);
        process(this, annotations, source, target, true);
        target.AddRuntimeAnnotations(annotations);
    }

    private static RuntimeDbFunction Create(IDbFunction function, RuntimeModel runtimeModel)
        => new(
            function.ModelName,
            runtimeModel,
            function.ReturnType,
            function.Name,
            function.Schema,
            function.StoreType,
            function.MethodInfo,
            function.IsScalar,
            function.IsAggregate,
            function.IsNullable,
            function.IsBuiltIn,
            function.TypeMapping,
            function.Translation);

    /// <summary>
    ///     Updates the function annotations that will be set on the read-only object.
    /// </summary>
    /// <param name="annotations">The annotations to be processed.</param>
    /// <param name="function">The source function.</param>
    /// <param name="runtimeFunction">The target function that will contain the annotations.</param>
    /// <param name="runtime">Indicates whether the given annotations are runtime annotations.</param>
    protected virtual void ProcessFunctionAnnotations(
        Dictionary<string, object?> annotations,
        IDbFunction function,
        RuntimeDbFunction runtimeFunction,
        bool runtime)
    {
    }

    private static RuntimeDbFunctionParameter Create(IDbFunctionParameter parameter, RuntimeDbFunction runtimeFunction)
        => runtimeFunction.AddParameter(
            parameter.Name,
            parameter.ClrType,
            parameter.PropagatesNullability,
            parameter.StoreType,
            parameter.TypeMapping);

    /// <summary>
    ///     Updates the function parameter annotations that will be set on the read-only object.
    /// </summary>
    /// <param name="annotations">The annotations to be processed.</param>
    /// <param name="parameter">The source function parameter.</param>
    /// <param name="runtimeParameter">The target function parameter that will contain the annotations.</param>
    /// <param name="runtime">Indicates whether the given annotations are runtime annotations.</param>
    protected virtual void ProcessFunctionParameterAnnotations(
        Dictionary<string, object?> annotations,
        IDbFunctionParameter parameter,
        RuntimeDbFunctionParameter runtimeParameter,
        bool runtime)
    {
    }

    private static RuntimeSequence Create(ISequence sequence, RuntimeModel runtimeModel)
        => new(
            sequence.Name,
            runtimeModel,
            sequence.Type,
            sequence.Schema,
            sequence.StartValue,
            sequence.IncrementBy,
            sequence.IsCyclic,
            sequence.MinValue,
            sequence.MaxValue);

    /// <summary>
    ///     Updates the sequence annotations that will be set on the read-only object.
    /// </summary>
    /// <param name="annotations">The annotations to be processed.</param>
    /// <param name="sequence">The source sequence.</param>
    /// <param name="runtimeSequence">The target sequence that will contain the annotations.</param>
    /// <param name="runtime">Indicates whether the given annotations are runtime annotations.</param>
    protected virtual void ProcessSequenceAnnotations(
        Dictionary<string, object?> annotations,
        ISequence sequence,
        RuntimeSequence runtimeSequence,
        bool runtime)
    {
    }

    /// <summary>
    ///     Updates the property annotations that will be set on the read-only object.
    /// </summary>
    /// <param name="annotations">The annotations to be processed.</param>
    /// <param name="property">The source property.</param>
    /// <param name="runtimeProperty">The target property that will contain the annotations.</param>
    /// <param name="runtime">Indicates whether the given annotations are runtime annotations.</param>
    protected override void ProcessPropertyAnnotations(
        Dictionary<string, object?> annotations,
        IProperty property,
        RuntimeProperty runtimeProperty,
        bool runtime)
    {
        base.ProcessPropertyAnnotations(annotations, property, runtimeProperty, runtime);

        if (runtime)
        {
            annotations.Remove(RelationalAnnotationNames.TableColumnMappings);
            annotations.Remove(RelationalAnnotationNames.ViewColumnMappings);
            annotations.Remove(RelationalAnnotationNames.SqlQueryColumnMappings);
            annotations.Remove(RelationalAnnotationNames.FunctionColumnMappings);
            annotations.Remove(RelationalAnnotationNames.DefaultColumnMappings);
        }
        else
        {
            annotations.Remove(RelationalAnnotationNames.ColumnOrder);
            annotations.Remove(RelationalAnnotationNames.Comment);
            annotations.Remove(RelationalAnnotationNames.Collation);

            if (annotations.TryGetValue(RelationalAnnotationNames.RelationalOverrides, out var overrides))
            {
                var runtimePropertyOverrides = new SortedDictionary<StoreObjectIdentifier, object>();
                foreach (var (storeObjectIdentifier, value) in (SortedDictionary<StoreObjectIdentifier, object>?)overrides!)
                {
                    var runtimeOverrides = Create((IRelationalPropertyOverrides)value, runtimeProperty);
                    runtimePropertyOverrides[storeObjectIdentifier] = runtimeOverrides;

                    CreateAnnotations(
                        (IRelationalPropertyOverrides)value, runtimeOverrides,
                        static (convention, annotations, source, target, runtime) =>
                            convention.ProcessPropertyOverridesAnnotations(annotations, source, target, runtime));
                }

                annotations[RelationalAnnotationNames.RelationalOverrides] = runtimePropertyOverrides;
            }
        }
    }

    private static RuntimeRelationalPropertyOverrides Create(
        IRelationalPropertyOverrides propertyOverrides,
        RuntimeProperty runtimeProperty)
        => new(
            runtimeProperty,
            propertyOverrides.ColumnNameOverridden,
            propertyOverrides.ColumnName);

    /// <summary>
    ///     Updates the relational property overrides annotations that will be set on the read-only object.
    /// </summary>
    /// <param name="annotations">The annotations to be processed.</param>
    /// <param name="propertyOverrides">The source relational property overrides.</param>
    /// <param name="runtimePropertyOverrides">The target relational property overrides that will contain the annotations.</param>
    /// <param name="runtime">Indicates whether the given annotations are runtime annotations.</param>
    protected virtual void ProcessPropertyOverridesAnnotations(
        Dictionary<string, object?> annotations,
        IRelationalPropertyOverrides propertyOverrides,
        RuntimeRelationalPropertyOverrides runtimePropertyOverrides,
        bool runtime)
    {
    }

    /// <summary>
    ///     Updates the key annotations that will be set on the read-only object.
    /// </summary>
    /// <param name="annotations">The annotations to be processed.</param>
    /// <param name="key">The source key.</param>
    /// <param name="runtimeKey">The target key that will contain the annotations.</param>
    /// <param name="runtime">Indicates whether the given annotations are runtime annotations.</param>
    protected override void ProcessKeyAnnotations(
        IDictionary<string, object?> annotations,
        IKey key,
        RuntimeKey runtimeKey,
        bool runtime)
    {
        base.ProcessKeyAnnotations(annotations, key, runtimeKey, runtime);

        if (runtime)
        {
            annotations.Remove(RelationalAnnotationNames.UniqueConstraintMappings);
        }
    }

    /// <summary>
    ///     Updates the index annotations that will be set on the read-only object.
    /// </summary>
    /// <param name="annotations">The annotations to be processed.</param>
    /// <param name="index">The source index.</param>
    /// <param name="runtimeIndex">The target index that will contain the annotations.</param>
    /// <param name="runtime">Indicates whether the given annotations are runtime annotations.</param>
    protected override void ProcessIndexAnnotations(
        Dictionary<string, object?> annotations,
        IIndex index,
        RuntimeIndex runtimeIndex,
        bool runtime)
    {
        base.ProcessIndexAnnotations(annotations, index, runtimeIndex, runtime);

        annotations.Remove(runtime ? RelationalAnnotationNames.TableIndexMappings : RelationalAnnotationNames.Filter);
    }

    /// <summary>
    ///     Updates the foreign key annotations that will be set on the read-only object.
    /// </summary>
    /// <param name="annotations">The annotations to be processed.</param>
    /// <param name="foreignKey">The source foreign key.</param>
    /// <param name="runtimeForeignKey">The target foreign key that will contain the annotations.</param>
    /// <param name="runtime">Indicates whether the given annotations are runtime annotations.</param>
    protected override void ProcessForeignKeyAnnotations(
        Dictionary<string, object?> annotations,
        IForeignKey foreignKey,
        RuntimeForeignKey runtimeForeignKey,
        bool runtime)
    {
        base.ProcessForeignKeyAnnotations(annotations, foreignKey, runtimeForeignKey, runtime);

        if (runtime)
        {
            annotations.Remove(RelationalAnnotationNames.ForeignKeyMappings);
        }
    }

    private static RuntimeTrigger Create(ITrigger trigger, RuntimeEntityType runtimeEntityType)
        => new(runtimeEntityType, trigger.ModelName, trigger.Name, trigger.TableName, trigger.TableSchema);

    /// <summary>
    ///     Updates the function annotations that will be set on the read-only object.
    /// </summary>
    /// <param name="annotations">The annotations to be processed.</param>
    /// <param name="trigger">The source trigger.</param>
    /// <param name="runtimeTrigger">The target trigger that will contain the annotations.</param>
    /// <param name="runtime">Indicates whether the given annotations are runtime annotations.</param>
    protected virtual void ProcessTriggerAnnotations(
        Dictionary<string, object?> annotations,
        ITrigger trigger,
        RuntimeTrigger runtimeTrigger,
        bool runtime)
    {
    }
}
